///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2011) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

// MakeDepend: library OMP
// MakeDepend: cflags OMP_FLAGS
#include <omp.h>

#include "CoreTools.h"
#include "VecUtilities.h"
#include "ParallelVecUtilities.h"
#include "lookup/LookAlign.h" 
#include "efasta/EfastaTools.h"
#include "paths/AssemblyCleanupTools.h"
#include "pairwise_aligners/SmithWatBandedA.h"

Bool scompare( const superb& s1, const superb& s2 ){
  int l1 = s1.FullLength();
  int l2 = s2.FullLength();
  return (l1 > l2);
}


Assembly::Assembly( const String in_superb_file, const String in_efasta_file ){
  // Loading scaffolds
  cout << Date( ) << ": loading superb file" << endl;
  ReadSuperbs( in_superb_file, scaffolds );
  
  // reading contig information
  LoadEfastaIntoStrings(in_efasta_file, efastas);
  tigMap.resize( efastas.size() );
  for ( int tid = 0; tid < efastas.isize(); tid++ )
    tigMap[tid] = ToString(tid);

  scaffMap.resize( scaffolds.size() );
  for ( int sid = 0; sid < scaffolds.isize(); sid++ )
    scaffMap[sid] = ToString(sid);
}


Assembly::Assembly( const vec<superb>& scaffoldsIn, const vec<efasta>& efastasIn ){

  scaffolds = scaffoldsIn;
  efastas   = efastasIn;
  tigMap.resize( efastas.size() );
  for ( int tid = 0; tid < efastas.isize(); tid++ )
    tigMap[tid] = ToString(tid);

  scaffMap.resize( scaffolds.size() );
  for ( int sid = 0; sid < scaffolds.isize(); sid++ )
    scaffMap[sid] = ToString(sid);
}


size_t Assembly::scaffoldsTotLen() const{
  size_t len = 0;
  for ( size_t is = 0; is < scaffolds.size(); is++ )
    len += scaffolds[is].FullLength();
  return len;
}

size_t Assembly::scaffoldsRedLen() const{
  size_t len = 0;
  for ( size_t is = 0; is < scaffolds.size(); is++ )
    len += scaffolds[is].ReducedLength();
  return len;
}

size_t Assembly::scaffoldsNtigs() const{
  size_t ntigs = 0;
  for ( size_t is = 0; is < scaffolds.size(); is++ )
    ntigs += scaffolds[is].Ntigs();
  return ntigs;
}

// check integrity of scafolds and contigs data: contig size in superb == contig size in efasta,
//  each contig used once and only once
void Assembly::check_integrity() const{
  
  cout << Date() << ": checking integrity" << endl;
  vec<int> tigs_used( efastas.size(), 0);
  for ( size_t i = 0; i < efastas.size( ); i++ )
  {    vec<String> s(1);
       s[0] = efastas[i];
       ValidateEfastaRecord(s);    }
  for ( size_t si = 0; si < scaffolds.size(); si++ ){
    const superb & s = scaffolds[si];
    ForceAssertGt( s.Ntigs(), 0 );
    for ( int tpos = 0; tpos < s.Ntigs(); tpos++ ){
      size_t tid = s.Tig(tpos);
      ForceAssertLt( tid, efastas.size() );
      if ( efastas[tid].Length1() != s.Len(tpos) ){
	DPRINT5( si, tpos, tid, s.Len(tpos), efastas[tid].Length1() );
	ForceAssertEq( efastas[tid].Length1(), s.Len(tpos) );
      }
      tigs_used[tid]++;
    }
  }
  vec<size_t> unused_tigs, overused_tigs;
  for ( size_t tid = 0; tid < tigs_used.size(); tid++ ){
    if ( tigs_used[tid] == 0 )
      unused_tigs.push_back( tid );
    else if ( tigs_used[tid] > 1 )
      overused_tigs.push_back(tid);
    
  }
  
  if ( unused_tigs.size() > 0 || overused_tigs.size() > 0 ){
    
    if ( unused_tigs.size() > 0 ){
      int max_un_size = efastas.at( unused_tigs[0] ).Length1();
      for ( size_t i = 0; i < unused_tigs.size(); i++ )
	if (  efastas.at( unused_tigs[i] ).Length1() > max_un_size )
	  max_un_size = efastas.at( unused_tigs[i] ).Length1();

      cout << Date() << ": maximum size of unused contig is : " << max_un_size << endl;
    }

    DPRINT2( unused_tigs.size(), overused_tigs.size() );
    ForceAssert( unused_tigs.size() == 0 && overused_tigs.size() == 0 );
  }
  
  return;
}


void Assembly::remove_small_scaffolds(const int min_scaffold_size) {
  cout << Date() << ": removing small scaffolds: " << endl;
  cout << Date() << ": initial number of scaffolds = " << scaffolds.size() << endl;
  for ( int si = 0; si < scaffolds.isize(); si++ )
    if ( scaffolds.at(si).ReducedLength() < min_scaffold_size ){
      scaffolds.erase( scaffolds.begin() + si );
      scaffMap.erase( scaffMap.begin() + si );
      si--;      
    }
  for ( int si = 0; si < scaffolds.isize(); si++ )
    ForceAssertGe( scaffolds.at(si).ReducedLength(), min_scaffold_size );
  cout << Date() << ": final number of scaffolds = " << scaffolds.size() << endl;
}

void Assembly::remove_contigs( const vec<Bool>& to_remove )
{
  vec<int> offsets( efastas.size(), 0 );

  int offset = 0;
  for ( size_t tid = 0; tid < efastas.size(); tid++ ){
    if ( to_remove[tid] ){
      offsets[tid] = -1;
      offset++;
    }
    else{ offsets[tid] = offset; }
  }
    
  ForceAssertEq( offsets.size(), efastas.size() );
  for ( int tid = 0; tid < offsets.isize(); tid++ ){
    if ( offsets[tid] > 0 ){
      efastas[ tid - offsets[tid] ] = efastas[tid];
      tigMap[tid - offsets[tid] ] = tigMap[tid];
    }      
  }
  efastas.resize( efastas.size() - offset );
  tigMap.resize( tigMap.size() - offset );
  ForceAssertEq( efastas.size(), tigMap.size() );
  
  for ( size_t si = 0; si < scaffolds.size(); si++ ){
    for ( int tpos = 0; tpos < scaffolds[si].Ntigs(); tpos++ ){
      int tid = scaffolds[si].Tig(tpos);
      if ( offsets[tid] >= 0 ){
	int newtid = tid - offsets[tid];
	ForceAssertGe( newtid, 0 );
	scaffolds[si].SetTig( tpos, newtid );
      }
      else{
	scaffolds[si].RemoveTigByPos( tpos );
	tpos--;
      }
    }
  }
}

   
void Assembly::remove_small_contigs( const int min_size_solo, const int min_size_in ){
  cout << Date() << ": removing small contigs and renumbering" << endl;
  vec<int> offsets( efastas.size(), 0 );
  vec<Bool> tigsToRemove( efastas.size(), False);
  for ( size_t si = 0; si < scaffolds.size(); si++ ){
    if ( scaffolds[si].Ntigs() == 1 ){
      if ( scaffolds[si].Len(0) < min_size_solo )
	tigsToRemove[ scaffolds[si].Tig(0) ] = True;
    }
    else{
      for ( int tpos = 0; tpos < scaffolds[si].Ntigs(); tpos++ ){
	if ( scaffolds[si].Len(tpos) < min_size_in )
	  tigsToRemove[ scaffolds[si].Tig(tpos) ] = True;
      }
    }
  }
  remove_contigs(tigsToRemove);
}

void Assembly::remove_unused_contigs(){
  cout << Date() << ": removing unused contigs and renumbering" << endl;
   vec<int> tigs_used( efastas.size(), 0);
  for ( size_t si = 0; si < scaffolds.size(); si++ ){
    const superb & s = scaffolds[si];
    ForceAssertGt( s.Ntigs(), 0 );
    for ( int tpos = 0; tpos < s.Ntigs(); tpos++ ){
      size_t tid = s.Tig(tpos);
      ForceAssertLt( tid, efastas.size() );
      if ( efastas[tid].Length1() != s.Len(tpos) ){
	DPRINT5( si, tpos, tid, s.Len(tpos), efastas[tid].Length1() );
	ForceAssertEq( efastas[tid].Length1(), s.Len(tpos) );
      }
      tigs_used[tid]++;
    }
  }

  size_t unusedCt = 0;
  vec<int> offsets( efastas.size(), 0 );
  int offset = 0;
  for ( size_t tid = 0; tid < efastas.size(); tid++ ){
    if ( ! tigs_used[tid] ){
      offsets[tid] = -1;
      offset++;
      unusedCt++;
    }
    else{ offsets[tid] = offset; }
  }
  cout << Date( ) << ": found " << unusedCt << " unused contigs, removing" << endl;
  ForceAssertEq( offsets.size(), efastas.size() );
  for ( int tid = 0; tid < offsets.isize(); tid++ ){
    if ( offsets[tid] > 0 ){
      efastas[ tid - offsets[tid] ] = efastas[tid];
      tigMap[tid - offsets[tid] ] = tigMap[tid];
    }      
  }
  efastas.resize( efastas.size() - offset );
  tigMap.resize( tigMap.size() - offset );
  ForceAssertEq( efastas.size(), tigMap.size() );
  
  cout << Date() << ": updating scaffolds tig ids" << endl;
  for ( size_t si = 0; si < scaffolds.size(); si++ ){
    for ( int tpos = 0; tpos < scaffolds[si].Ntigs(); tpos++ ){
      int tid = scaffolds[si].Tig(tpos);
      if ( offsets[tid] >= 0 ){
	int newtid = tid - offsets[tid];
	ForceAssertGe( newtid, 0 );
	scaffolds[si].SetTig( tpos, newtid );
      }
      else{
	scaffolds[si].RemoveTigByPos( tpos );
	tpos--;
      }
    }
  }
  cout << Date() << ": done with removing unused contigs" << endl;
}


void Assembly::reorder(){
  // Sorting scaffolds according to size and renumbering contigs according
  //   to sequential appearance in scaffolds
  cout << Date() << ": sorting scaffolds" << endl;
  SortSync( scaffolds, scaffMap, scompare );
  renumber();
}



// renumber all the contigs sequentially according to the scaffold
void Assembly::renumber(){
  cout << Date() << ": renumbering contigs for ordered scaffolds" << endl;
  vec<String> otigMap;  
  vec<superb> oscaffolds = scaffolds;
  vec<efasta> oefastas;
  int cTid = -1;
  for ( size_t is = 0; is < scaffolds.size(); is++ ){
    for ( int tpos = 0; tpos < scaffolds[is].Ntigs(); tpos++ ){
      cTid++;
      int oTid = scaffolds[is].Tig(tpos);
      oscaffolds.at(is).SetTig( tpos, cTid );
      oefastas.push_back( efastas.at(oTid) );
      otigMap.push_back( tigMap.at(oTid) );
    }
  }
  efastas.resize(0);
  tigMap.resize(0);
  size_t newScaffoldsTotLen = 0, newScaffoldsRedLen = 0;
  for ( size_t is = 0; is < scaffolds.size(); is++ ){
    newScaffoldsTotLen += scaffolds[is].FullLength();
    newScaffoldsRedLen += scaffolds[is].ReducedLength();
  }
  scaffolds = oscaffolds; 
  oscaffolds.resize(0);
  efastas = oefastas; 
  oefastas.resize(0);
  tigMap = otigMap;
}

struct sData{
  
  sData() : AmbEventCount_(0), MaxAmbCount_(0), RedLen_(0), MinRedLen_(0), MaxRedLen_(0), 
	    FullLen_(0), MinFullLen_(0), MaxFullLen_(0) {};

  int AmbEventCount_, MaxAmbCount_;
  int RedLen_, MinRedLen_, MaxRedLen_;
  int FullLen_, MinFullLen_, MaxFullLen_;
  basevector seq_, minseq_, maxseq_;
  basevector seq_rc_, minseq_rc_, maxseq_rc_;
};

void Assembly::dedup( const Bool Exact ) {
  // Remove duplicate scaffolds.
  cout << Date() << ": removing duplicate scaffolds: " << endl;
  cout << Date() << ": initial number of scaffolds = " << scaffolds.size() << endl;

  // cononicalize efastas
  vec<efasta> cefastas( efastas.size() );
  vec<efasta> cefastas_rc( efastas.size() );
  for ( size_t ci = 0; ci < efastas.size(); ci++ ){
    efastas[ci].Canonicalize( cefastas[ci] );
    ValidateEfastaRecord( cefastas[ci] );
    efasta eseq_rc = efastas[ci];
    eseq_rc.ReverseComplement();
    eseq_rc.Canonicalize( cefastas_rc[ci] );
    ValidateEfastaRecord( cefastas_rc[ci] );
  }

  // compute scaffold data

  vec< vec< vec<basevector> > > s2altTigs( scaffolds.size() );
  vec<sData> s2data( scaffolds.size() );

  for ( int si = 0; si < scaffolds.isize(); si++ ){
    s2altTigs[si].resize( scaffolds[si].Ntigs() );
    for ( int i = 0; i < scaffolds[si].Ntigs(); i++ ){
      int ti = scaffolds[si].Tig(i);
      efasta& tigi = cefastas[ ti ];
      s2data[si].AmbEventCount_ += tigi.AmbEventCount();
      s2data[si].MaxAmbCount_   += tigi.AmbCount();
      s2data[si].MinRedLen_     += tigi.MinLength();
      s2data[si].MaxRedLen_     += tigi.MaxLength();
      s2data[si].MinFullLen_    += tigi.MinLength();
      s2data[si].MaxFullLen_    += tigi.MaxLength();
      s2data[si].RedLen_        += tigi.Length1();
      s2data[si].FullLen_       += tigi.Length1();
      if ( i < scaffolds[si].Ntigs() -1 ){
	s2data[si].MinFullLen_ += scaffolds[si].Gap(i) - scaffolds[si].Dev(i);
	s2data[si].MaxFullLen_ += scaffolds[si].Gap(i) + scaffolds[si].Dev(i);
	s2data[si].FullLen_    += scaffolds[si].Gap(i);
      }

      fastavector fseq;
      tigi.FlattenMinTo( fseq );
      s2altTigs[si][i].push_back( fseq.ToBasevector() );
      if ( tigi.Ambiguities() > 0 ){
	tigi.FlattenTo( fseq );
	s2altTigs[si][i].push_back( fseq.ToBasevector() );
	tigi.FlattenMaxTo( fseq );
	s2altTigs[si][i].push_back( fseq.ToBasevector() );
      }else{
	s2altTigs[si][i].push_back( fseq.ToBasevector() );
	s2altTigs[si][i].push_back( fseq.ToBasevector() );
      }
    }
    DPRINT3( si, s2data[si].AmbEventCount_, s2data[si].MaxAmbCount_ );
  }
  
  for ( int si = 0; si < scaffolds.isize(); si++ ) {
    int Ni = scaffolds[si].Ntigs();
    basevector seq, minseq, maxseq;
    for ( int ci = 0; ci < Ni; ci++ ){
      minseq = Cat( minseq, s2altTigs[si][ci][0] );
      seq    = Cat( seq, s2altTigs[si][ci][1] );
      maxseq = Cat( maxseq, s2altTigs[si][ci][2] );
      if ( ci < Ni -1 )
	for ( int pi = 0; pi < scaffolds[si].Gap(ci); pi++ ){
	  seq.push_back(0); minseq.push_back(0); maxseq.push_back(0);
	}
    }
    s2data[si].seq_ = seq; s2data[si].minseq_ = minseq; s2data[si].maxseq_ = maxseq; 
    basevector seq_rc, minseq_rc, maxseq_rc;
    for ( int i = 0; i < Ni; i++ ){
      int ci = Ni -1 -i;
      basevector ci_minseq_rc = s2altTigs[si][ci][0]; ci_minseq_rc.ReverseComplement();
      basevector ci_seq_rc = s2altTigs[si][ci][1];    ci_seq_rc.ReverseComplement();
      basevector ci_maxseq_rc = s2altTigs[si][ci][2]; ci_maxseq_rc.ReverseComplement();
      minseq_rc = Cat( minseq_rc, ci_minseq_rc );
      seq_rc    = Cat( seq_rc, ci_seq_rc );
      maxseq_rc = Cat( maxseq_rc, ci_maxseq_rc );
      if ( ci > 0 )
	for ( int pi = 0; pi < scaffolds[si].Gap(ci-1); pi++ ){
	  seq_rc.push_back(0); minseq_rc.push_back(0); maxseq_rc.push_back(0);
	}
    }
    ForceAssertEq( minseq.size(), minseq_rc.size() );
    ForceAssertEq( seq.size(), seq_rc.size() );
    ForceAssertEq( maxseq.size(), maxseq_rc.size() );
    s2data[si].seq_rc_ = seq_rc; s2data[si].minseq_rc_ = minseq_rc; s2data[si].maxseq_rc_ = maxseq_rc; 
  }


  int removed_count = 0;
  int removed_size  = 0;
  vec <Bool> to_remove ( scaffolds.size(), False );

  for ( int si = 0; si < scaffolds.isize(); si++ ) {
    if ( to_remove[si] ) continue;
    int Ni = scaffolds[si].Ntigs();
    for ( int sj = si + 1; sj < scaffolds.isize(); sj++ ) {
      if ( to_remove[sj] ) continue;
      int Nj = scaffolds[sj].Ntigs();
      ForceAssertEq( Nj, s2altTigs[sj].isize() );
      
      // roughly check if scaffolds could be duplicates of each other 
      if ( Ni != Nj ) 
	continue;
      if ( s2data[si].FullLen_ != s2data[sj].FullLen_ && 
	   ( s2data[si].MaxFullLen_ < s2data[sj].MinFullLen_ || 
	     s2data[sj].MaxFullLen_ < s2data[si].MinFullLen_ ) )
	continue;
      if ( s2data[si].RedLen_ != s2data[sj].RedLen_ && 
	   ( s2data[si].MaxRedLen_ < s2data[sj].MinRedLen_ || 
	     s2data[sj].MaxRedLen_ < s2data[si].MinRedLen_ ) )
	continue;
      
      cout << "\n\n\n";
      cout << Date() << ":  ---------     comparing scaffolds:" << endl;
      DPRINT2( si, sj );
      DPRINT2( s2data[si].AmbEventCount_, s2data[sj].AmbEventCount_ );
      DPRINT2( s2data[si].MaxAmbCount_, s2data[sj].MaxAmbCount_ );
      DPRINT2( s2data[si].seq_.size(), s2data[sj].seq_.size());

      // more detailed check      
      vec<int> areRcs( Ni, 0 );
      vec<int> areEqual( Ni, 0 );
      for ( int ci = 0; ci < Ni; ci++ ){
	
	if ( ci > 0 &&  areEqual[ci-1] != 1 &&  areRcs[ci-1] != 1 ) continue;
	
	int ti = scaffolds[si].Tig(ci);
	vec< vec<String> > iblocks;
	cefastas[ti].GetBlocks( iblocks );

	if ( ci == 0 || areEqual[ci-1] == 1 ){
	  // checking duplication
	  int cje = ci;
	  int tje = scaffolds[sj].Tig(cje);
	  if ( cefastas[ti] == cefastas[tje] ){
	    areEqual[ci] = 1;
	  }else if ( cefastas[ti].AmbEventCount() == cefastas[tje].AmbEventCount() ){
	    vec< vec<String> > eblocks;
	    cefastas[tje].GetBlocks( eblocks );
	    Bool pathEqual = True;
	    for ( int ib = 0; ib < iblocks.isize(); ib++ ){
	      if ( iblocks[ib].size() == 1 && eblocks[ib].size() == 1 ){
		if ( iblocks[ib][0] != eblocks[ib][0] ){
		  pathEqual =False;
		  break;
		}
	      }else{
		Bool foundEqualRoute = False;
		for ( int ir1 = 0; ir1 < iblocks[ib].isize(); ir1++ )
		  for ( int ir2 = 0; ir2 < eblocks[ib].isize(); ir2++ )
		    if ( iblocks[ib][ir1] == eblocks[ib][ir2] ){
		      foundEqualRoute = True;
		      break;
		    }
		if ( ! foundEqualRoute ){
		  pathEqual = False;
		  break;
		}
	      }
	    }
	    if ( pathEqual ) areEqual[ci] = 1;
	  }
	}
	
	if ( ci == 0 || areRcs[ci-1] == 1 ){
	  // checking reverse duplication	  
	  int cjr = Nj -ci -1;
	  int tjr = scaffolds[sj].Tig(cjr);
	  efasta eseq_rc = cefastas_rc[tjr];
	  if ( cefastas[ti] == eseq_rc ){
	    areRcs[ci] = 1;
	  }else if ( cefastas[ti].AmbEventCount() == eseq_rc.AmbEventCount() ){
	    vec< vec<String> > rblocks;
	    eseq_rc.GetBlocks( rblocks );
	    Bool pathEqual = True;
	    for ( int ib = 0; ib < iblocks.isize(); ib++ ){
	      if ( iblocks[ib].size() == 1 && rblocks[ib].size() == 1 ){
		if ( iblocks[ib][0] != rblocks[ib][0] ){
		  pathEqual =False;
		  break;
		}
	      }else{
		Bool foundEqualRoute = False;
		for ( int ir1 = 0; ir1 < iblocks[ib].isize(); ir1++ )
		  for ( int ir2 = 0; ir2 < rblocks[ib].isize(); ir2++ )
		    if ( iblocks[ib][ir1] == rblocks[ib][ir2] ){
		      foundEqualRoute = True;
		      break;
		    }
		if ( ! foundEqualRoute ){
		  pathEqual = False;
		  break;
		}
	      }
	    }
	    if ( pathEqual ) areRcs[ci] = 1;
	  }
	}
      }    
      
      if ( Sum( areRcs ) > 0 || Sum( areEqual ) > 0 )
	DPRINT4( areRcs.size(), Sum( areRcs ), areEqual.size(), Sum( areEqual ) );

      if ( Sum( areRcs ) == areRcs.isize() || Sum( areEqual ) == areEqual.isize() ){

	String type = Sum( areRcs ) == areRcs.isize() ? "reverse complement" : "duplicate";
	
	cout << Date() << ": scaffold " << sj << "(l=" << scaffolds[si].FullLength() << ") is " << type << " of " << si
	     << " (length " << scaffolds[si].FullLength() << ")" << endl;
	
	to_remove[sj] = True;
	removed_count++;
	removed_size += scaffolds[sj].ReducedLength();
	
      }else if ( ! Exact) { 
	cout << "\n";
	cout << Date() << ": ------------- RUNNING BANDED SMITH-WATERMAN -------------------\n\n";
	int amb_diff_count = Max( s2data[si].maxseq_.isize() - s2data[si].minseq_.isize(), 
				  s2data[sj].maxseq_.isize() - s2data[sj].minseq_.isize() );
	int amb_event_count = Max( s2data[si].AmbEventCount_, s2data[sj].AmbEventCount_ );
	int amb_max_count   = Max( s2data[si].MaxAmbCount_, s2data[sj].MaxAmbCount_ );
	DPRINT3( amb_diff_count, amb_event_count, amb_max_count );
	// no exact match, check alignment
	for ( int iter = 0; iter < 3; iter++ ){
	  if ( to_remove[sj] ) continue;
	  basevector si_seq;  
	  if ( iter == 0 )      si_seq = s2data[si].seq_;
	  else if ( iter == 1 ) si_seq = s2data[si].minseq_;
	  else if ( iter == 2 ) si_seq = s2data[si].maxseq_;
	  basevector sj_seq;
	  if ( iter == 0 )      sj_seq = s2data[sj].seq_;
	  else if ( iter == 1 ) sj_seq = s2data[sj].minseq_;
	  else if ( iter == 2 ) sj_seq = s2data[sj].maxseq_;

	  // flank sequences with the same ends for the case where efasta amibugity is at the edge
	  Bool FlankSequences = False;
	  if ( efastas[ scaffolds[si].Tig(0) ].front() == '{' || efastas[ scaffolds[sj].Tig(0) ].front() == '{' ||
	       efastas[ scaffolds[si].Tig(0) ].back() == '}' || efastas[ scaffolds[sj].Tig(0) ].back() == '}' )
	    FlankSequences = True;

	  basevector flank;
	  flank.SetFromString("AAAAAAAAAA");
	  if ( FlankSequences ){
	    cout << Date()  << " Flanking" << endl;
	    si_seq = Cat( flank, si_seq, flank );
	    sj_seq = Cat( flank, sj_seq, flank );
	  }

	  align aF;
	  int min_overlap = Min( si_seq.size(), sj_seq.size() );
	  DPRINT( min_overlap );
	  int errorsF = 0;
	  int lenDiff = abs( si_seq.isize() - sj_seq.isize() );
	  int bandwidth = 2 * lenDiff; 
	  int maxErr = amb_max_count + amb_event_count + lenDiff + 1 + round( 0.0001 * (double)si_seq.isize() );
	  DPRINT( maxErr );
	  bandwidth = Max( bandwidth, 10 );
	  DPRINT( bandwidth );
	  cout << Date() << ": aligning forward" << endl;
	  float scoreF = 
	    SmithWatBandedA2<unsigned short>(si_seq, sj_seq, 0, bandwidth, aF, errorsF );  
	  look_align laF;
	  laF.ResetFromAlign( aF, si_seq, sj_seq );
	  DPRINT5( scoreF, laF.a.extent1(), laF.a.extent2(), errorsF, maxErr ); 
	  if ( ( laF.a.extent1() >= min_overlap || laF.a.extent2() >= min_overlap ) && 
	       errorsF <= maxErr && laF.Fw1() ){
	    cout << Date() << ": scaffold " << sj << "(l=" << scaffolds[si].FullLength() << ") is a copy of " << si
		 << " (length " << scaffolds[si].FullLength() << ")" << endl;
	    
	    to_remove[sj] = True;
	    removed_count++;
	    removed_size += scaffolds[sj].ReducedLength();
	  }else{
	    basevector sj_seq_rc;
	    if ( iter == 0 )      sj_seq_rc = s2data[sj].seq_rc_;
	    else if ( iter == 1 ) sj_seq_rc = s2data[sj].minseq_rc_;
	    else if ( iter == 2 ) sj_seq_rc = s2data[sj].maxseq_rc_;
	    
	    if ( FlankSequences )
	      sj_seq_rc = Cat( flank, si_seq, flank );

	    align aR;
	    int errorsR = 0;
	    cout << Date() << ": aligning reverse" << endl;
	    int scoreR = 
	      SmithWatBandedA2<unsigned short>(si_seq, sj_seq_rc, 0, bandwidth, aR, errorsR );  
	    look_align laR;
	    laR.ResetFromAlign( aR, si_seq, sj_seq_rc );
	    DPRINT5( scoreR, laR.a.extent1(), laR.a.extent2(), errorsR, maxErr ); 
	    if ( ( laR.a.extent1() >= min_overlap || laR.a.extent2() >= min_overlap ) && 
		 errorsR <= maxErr && laR.Fw1() ){
	      cout << Date() << ": scaffold " << sj << "(l=" << scaffolds[si].FullLength() << ") is reverse-complement of " << si
		   << " (length " << scaffolds[si].FullLength() << ")" << endl;
	      
	      to_remove[sj] = True;
	      removed_count++;
	      removed_size += scaffolds[sj].ReducedLength();
	    }
	  }
	  cout << Date() << ":  vvvvvvvvvvvvvv  DONE WIHT ALIGNMENT vvvvvvvvvvvvvvv\n\n\n" << endl;
	}
      }
    }
  }

  EraseIf( scaffolds, to_remove );
  EraseIf( scaffMap, to_remove );
  EraseIf( s2data, to_remove );
  EraseIf( s2altTigs, to_remove );

  cout << Date() << ": removed " << removed_count << " duplicate scaffolds"
       << " (" << removed_size << " bases)" << endl;
  cout << Date() << ": final number of scaffolds = " << scaffolds.size() << endl;
  remove_unused_contigs();
}


void Assembly::dedup2() {
  // Remove scaffolds that are possibly duplicate
  // The criteria are:
  // 1. two scaffolds have same number of contigs (n_contig)
  // 2. n_contig >= 2
  // 3. Each contig and gap much match 
  //    - Gaps are matched whan [ gap_size +/- 3 * std ] overlap
  //    - Contigs are matched when 
  //      - perfect efasta match for contig size < 50,000
  //      - 1/100 mismatch kmer rate for contig size >= 50,000 (!!!!! this is only meant to be temporary fix to the problem)
  // 4. Both rc and fw duplicates are checked

  int VERBOSITY = 0;
  const int EfastaMatchSize = 50 * 1000; 
  const int MaxDev  = 3;
  const double MaxMismatchRate = 0.01;
  cout << Date() << ": " << "Remove possible duplicate scaffolds" << endl;
  cout << Date() << ": " << "initial number of scaffolds = " << scaffolds.size() << endl;
  
  int removed_count = 0;
  int removed_size = 0;
  vec <Bool> to_remove ( scaffolds.size(), False );
  for ( int si = 0; si < scaffolds.isize(); si++ ) {
    if ( to_remove[si] ) continue;
    int Ni = scaffolds[si].Ntigs();
    //if ( Ni < 2 ) continue;
    for ( int sj = si + 1; sj < scaffolds.isize(); sj++ ) {
      if ( to_remove[sj] ) continue;
      int Nj = scaffolds[sj].Ntigs();
      if ( Nj != Ni ) continue;
      //VERBOSITY = ( si == 1 && sj == 6 ? 1: 0 );
      // check scaffolds duplicate in two passes: fw in pass=0, rc in pass=1
      for( int pass = 0; pass < 2; pass++ ) {
	if ( VERBOSITY >= 1 )
	cout << Date() << ": " << "pass= " << pass << endl;
	if ( VERBOSITY >= 1 )
	  cout << Date() << ": " << "check gap " << endl ;
	// compare the gap of two scaffolds
	bool gap_match = true;
	for ( int igap = 0; igap < Ni-1; ++igap ) {
	  int gap1 = scaffolds[si].Gap(igap);
	  int dev1 = scaffolds[si].Dev(igap);
	  int gap2 = scaffolds[sj].Gap( pass == 0 ? igap : Ni -2 - igap);
	  int dev2 = scaffolds[sj].Dev( pass == 0 ? igap : Ni -2 - igap);
	  if ( IntervalOverlap( gap1 - MaxDev * dev1, gap1 + MaxDev * dev1 + 1,
		gap2 - MaxDev * dev2, gap2 + MaxDev * dev2 + 1) == 0 ) {
	    gap_match = false;
	    break;
	  }
	}
	if ( !gap_match) continue;
	if ( VERBOSITY >= 1 )
	  cout << Date() << ": " << "check contigs " << endl ;

	// check if the contigs matchs
	bool tig_match = true;
	if ( VERBOSITY >= 1 )
	  cout<< Date()  << " ncontigs= " << Ni << endl;
	for ( int itig = 0; itig < Ni; ++itig) {
	  if ( VERBOSITY >= 1 )
	    cout << Date() << ": check contig " << itig << ": "
	    << " vs " << scaffolds[sj].Tig( pass == 0 ? itig :  Nj -1 -itig ) << endl;
	  efasta &tigi = efastas[ scaffolds[si].Tig(itig) ];
	  efasta &tigj = efastas[ scaffolds[sj].Tig( pass == 0 ? itig :  Nj -1 -itig ) ] ;

	  if ( VERBOSITY >= 1 )
	    cout << Date() << ": tigi.size()= " << tigi.size() << endl;
	  if ( VERBOSITY >= 1 )
	    cout << Date() << ": tigj.size()= " << tigj.size() << endl;
	  
	  // check if contig sizes match
	  int tig_size = ( tigi.size() + tigj.size()  ) /2;
	  int MaxMismatch =  tig_size * MaxMismatchRate ;
	  if ( abs( (int)tigi.size() - (int)tigj.size() ) > MaxMismatch ) { tig_match = false; break; }

	  // require perfect efasta match if contig size less than EfastaMatchSize
	  if ( tig_size < EfastaMatchSize ) {
	    if ( VERBOSITY >= 1 )
	      cout << Date() << ":  check efasta " << endl;
	    vec<basevector> v_ibases, v_jbases;
	    tigi.ExpandTo(v_ibases);
	    tigj.ExpandTo(v_jbases);
	    if ( pass == 1 ) 
	      for ( size_t k = 0; k < v_jbases.size(); ++k ) { v_jbases[k].ReverseComplement(); }
	    bool foundEqual = False;
	    for ( size_t vi = 0; vi < v_ibases.size() && ! foundEqual; vi++ ){
	      for ( size_t vj = 0; vj < v_jbases.size(); vj++ ){
		if ( v_jbases[vj].size() != v_ibases[vi].size() )
		  continue;
		basevector jbases = v_jbases[vj];
		if ( jbases == v_ibases[vi] ){
		  foundEqual = True;
		  break;
		}
	      }
	    }
	    if ( ! foundEqual ) {
	      tig_match = false;
	      break;
	    }
	  } 
	  // larger contig size. do kmer matching
	  else {
	    if ( VERBOSITY >= 1 )
	      cout << Date() << ":  check kmers " << endl;
	    basevector base1, base2;
	    tigi.FlattenTo( base1 );
	    tigj.FlattenTo( base2 );
	    if ( pass == 1 ) base2.ReverseComplement();
	    const int K = 24;
	    ForceAssertGt( base1.isize( ), K );
	    vec< basevector > kmers1( base1.isize( ) - K + 1);
            #pragma omp parallel for
	    for ( int jz = 0; jz <= base1.isize( ) - K; jz += 1000 ) 
	      for ( int j = jz; j <= Min( base1.isize( ) - K, jz + 1000 ); j++ ) 
		kmers1[j].SetToSubOf( base1, j, K ); 
	      ParallelUniqueSort(kmers1);    
	    ForceAssertGt( base2.isize( ), K );
	    vec< basevector > kmers2( base2.isize( ) - K + 1);
            #pragma omp parallel for
	    for ( int jz = 0; jz <= base2.isize( ) - K; jz += 1000 ) 
	      for ( int j = jz; j <= Min( base2.isize( ) - K, jz + 1000 ); j++ ) 
	        kmers2[j].SetToSubOf( base2, j, K ); 
	    ParallelUniqueSort(kmers2);    

	    // compare how many kmers are identical for the two sorted list
	    int nkmer1 = kmers1.size(), nkmer2 = kmers2.size();
	    int nkmer = (nkmer1 + nkmer2)/2;
	    int count = 0;
	    for( size_t i = 0, j = 0; i < kmers1.size() && j < kmers2.size(); ) {
	      if ( kmers1[i] > kmers2[j] ) j++;
	      else if ( kmers1[i] < kmers2[j] )  i++;
	      else count++, i++, j++;
	    }
	    if ( VERBOSITY >= 1 ) {
	      cout << Date() << ": nkmer= " << nkmer << endl;
	      cout << Date() << ": duplicate= " << count << endl;
	    }
	    if ( abs(nkmer - count) > int( nkmer * MaxMismatchRate) ) {
	      tig_match = false;
	      break;
	    }
	  }
	} // for itig
	if ( !tig_match ) continue;
	// ---------------------------------------------------------------------------
	// Now we concluded that the two scaffolds are duplicate. Remove the later one
	// --------------------------------------------------------------------------
	{
	  String type = pass == 0 ? "fw duplicate" : "rc duplicate";
	  cout << Date() << ": scaffold " << sj << " is " << type << " of " << si
	    << " (length " << scaffolds[si].FullLength() << ")" << endl;
	  to_remove[sj] = True;
	  removed_count++;
	  removed_size += scaffolds[sj].ReducedLength();
	  break; // do not go second pass 
	}
      } // end pass 2
    } // end for sj
  } // end for si
  
  // now remove the duplicate scaffolds 
  {
    EraseIf( scaffolds, to_remove );
    EraseIf( scaffMap, to_remove );
  }

  cout << Date() << ": removed " << removed_count << " duplicate scaffolds"
       << " (" << removed_size << " bases)" << endl;
  cout << Date() << ": final number of scaffolds = " << scaffolds.size() << endl;
  remove_unused_contigs();
}

void Assembly::dedup_exact() {
  // Remove duplicate scaffolds...currently only handles case of singleton contigs
  // which are duplicates fw or rc. --bruce 8 Jun 2011
  cout << Date() << ": removing duplicate scaffolds: " << endl;
  cout << Date() << ": initial number of scaffolds = " << scaffolds.size() << endl;
  int removed_count = 0;
  int removed_size = 0;
  for ( int si = 0; si < scaffolds.isize(); si++ ) {
    if (scaffolds[si].Ntigs() != 1) continue;
    for ( int sj = si + 1; sj < scaffolds.isize(); sj++ ) {
      if (scaffolds[sj].Ntigs() != 1) continue;
      efasta &tigi = efastas[scaffolds[si].Tig(0)];
      efasta &tigj = efastas[scaffolds[sj].Tig(0)];
      if (tigi.size() != tigj.size()) continue;
      basevector ibases, jbases;
      tigi.FlattenTo(ibases);
      tigj.FlattenTo(jbases);
      bool rc = False;
      if (ibases != jbases) {
	jbases.ReverseComplement();
	rc = True;
	if (ibases != jbases) continue;
      }
      /*
      cout << "scaffold " << sj << " duplicate of " << si
	   << " (length " << scaffolds[si].FullLength()
	   << (rc ? " rc" : "") << ")" << endl;
      */
      scaffolds.erase( scaffolds.begin() + sj );
      scaffMap.erase( scaffMap.begin() + sj );
      sj--;
      removed_count++;
      removed_size += ibases.size();
    }
  }
  cout << Date() << ": removed " << removed_count << " duplicate scaffolds"
       << " (" << removed_size << " bases)" << endl;
  cout << Date() << ": final number of scaffolds = " << scaffolds.size() << endl;
  remove_unused_contigs();
}


void Assembly::Write( const String head_out ) const {

  // writing output
  cout << Date() << ": writing output files" << endl;
  WriteSuperbs( head_out + ".superb", scaffolds );
  WriteSummary( head_out + ".summary", scaffolds );


  
  Ofstream( efout, head_out + ".contigs.efasta" );
  for ( size_t id = 0; id < efastas.size(); id++ )
    efastas[id].Print(efout, "contig_" + ToString(id) );
}

void Assembly::WriteExtra( const String head_out ) const{

  vec<fastavector> fastas(efastas.size());
  for ( size_t id = 0; id < efastas.size(); id++ )
    efastas[id].FlattenTo( fastas[id] );
  Ofstream( fout, head_out + ".contigs.fasta" );
  for ( size_t id = 0; id < fastas.size(); id++ )
    fastas[id].Print(fout, "contig_" + ToString(id) );

  {
    vecfastavector vec_tigs;
    for ( size_t i = 0; i < fastas.size( ); i++ )
      vec_tigs.push_back_reserve( fastas[i] );
    vec_tigs.WriteAll( head_out + ".contigs.vecfasta" ); 
  }

  vecbasevector bases( efastas.size() );
  for ( size_t id = 0; id < efastas.size(); id++ )
    efastas[id].FlattenTo( bases[id] );
  
  bases.WriteAll( head_out + ".contigs.fastb" );
  
  Ofstream( cmout, head_out + ".contigs.mapping" );
  for ( size_t id = 0; id < tigMap.size(); id++ )
    cmout << ToString(id) + " from " + ToString( tigMap[id] ) << "\n";

  Ofstream( smout, head_out + ".superb.mapping" );
  for ( size_t is = 0; is < scaffMap.size(); is++ )
    smout << ToString(is) + " from " + ToString( scaffMap[is] ) << "\n";

  WriteScaffoldedEFasta( head_out + ".assembly.efasta", efastas, scaffolds );
  WriteScaffoldedFasta( head_out + ".assembly.fasta", fastas, scaffolds );

}
